@@ -249,12 +249,18 @@ sub construct_boolean($s, $n) {
 
 sub construct_call($s, $p) {
     my ($k, $v) = @$p;
-    "$k" =~ /^($sym):?$/ or die;
+    "$k" =~ /^($sym)($bp?)$/ or die;
     my $fn = $1;
+    my $args = $2;  # TODO add these args to value args
+    if ($args) {
+        $args =~ s/^\((.*)\)$/$1/ or die;
+    }
     $fn =~ s/^(let|try|catch)$/$1*/;
     $main_called = 1 if $fn eq 'main';
+    $args = 'val'->new(undef, $args);
+    $args->{ytag} = 'ysexpr';
     $v = SEQ($v) unless is_seq($v);
-    L(S($fn), map $s->construct($_), elems($v));
+    L(S($fn), map $s->construct($_), $args, elems($v));
 }
 
 sub construct_def($s, $p) {
@@ -292,17 +298,18 @@ sub get_sig {
 
 sub construct_defn($s, $p) {
     my ($k, $v) = @$p;
-    my ($def, $name, $args, $body) = $s->_defn_parse($k, $v);
+    my ($def, $name, $args, $body) = $s->_defn_parse($k, $v, 0);
     return L($def, $name, V(@$args), @$body);
 }
 
 sub construct_defn_multi($s, $p) {
     my ($k, $v) = @$p;
-    my $def = S('defn');
-    my $name = S($k);
+    text($k) =~ /^(defn|defmacro)\s+($sym)$/ or die;
+    my $def = $1;
+    my $name = S($2);
     my @defs = map {
         my ($k, $v) = @$_;
-        my ($def, undef, $args, $body) = $s->_defn_parse($k, $v);
+        my (undef, undef, $args, $body) = $s->_defn_parse($k, $v, 1);
         L(V(@$args), @$body);
     } pairs($v);
     return L($def, $name, @defs);
@@ -310,19 +317,22 @@ sub construct_defn_multi($s, $p) {
 
 sub construct_fn($s, $p) {
     my ($k, $v) = @$p;
-    my ($def, $name, $args, $body) = $s->_defn_parse($k, $v);
+    my ($def, $name, $args, $body) = $s->_defn_parse($k, $v, 0);
     return L(FN, V(@$args), @$body);
 }
 
-sub _defn_parse($s, $k, $v) {
-    my $macro = 0;
-    text($k) =~ /^($sym?)?\((.*)\)$/ or die;
-    my $name = S($1);
-    my $sig = $2;
-    if ($sig =~ s/^\(//) {
-        die "Bad defmacro syntax '$k'"
-            unless $sig =~ s/\)$//;
-        $macro = 1;
+sub _defn_parse($s, $k, $v, $m) {
+    my ($def, $name, $sig);
+    if ($m) {
+        text($k) =~ /^($sym?)?\((.*)\)$/ or XXX $k;
+        $def = '';
+        $name = S($1);
+        $sig = $2;
+    } else {
+        text($k) =~ /^(fn|defn|defmacro)\s+($sym?)?\((.*)\)$/ or XXX $k;
+        $def = S($1);
+        $name = S($2);
+        $sig = $3;
     }
     my ($args, $dargs) = get_sig($sig);
     my $defn = L( DEF, $name, L( FN, L, nil ) );
@@ -333,7 +343,6 @@ sub _defn_parse($s, $k, $v) {
             ? ($s->construct_let($seq, $args, $dargs))
             : map $s->construct($_), @{$seq->elem},
     ];
-    my $def = $macro ? S('defmacro') : S('defn');
     return $def, $name, $args, $body;
 }
 
@@ -500,6 +509,19 @@ sub construct_val($s, $n) {
     T("$n");
 }
 
+sub construct_when($s, $p) {
+    my ($k, $v) = @$p;
+    (my $expr = "$k") =~ s/ ?([?|])$// or die;
+    my $fn = $1 eq '?' ? 'when' : 'when-not';
+    my $cond = read_ysexpr($expr);
+    my @elems = is_seq($v) ? elems($v) : $v;
+    L(
+        S($fn),
+        $cond,
+        map $s->construct($_), @elems,
+    );
+}
+
 sub construct_ysexpr($s, $n) {
     read_ysexpr($n);
 }
@@ -535,7 +557,7 @@ sub need_main_call($ast) {
 #------------------------------------------------------------------------------
 
 my $dyn = qr<(?:\*$sym\*)>;
-my $op = qr{(?:[-+*/]|[<>=]=?|and|or)};
+my $op = qr{(?:[-+*/]|[<>=]=?|and|or|\.\.)};
 
 my $pn = qr=(?:->|~@|[\'\`\[\]\{\}\(\)\~\^\@])=;
 # my $pn = qr<(?:~@|[\'\`\[\]\{\}\(\)\~\^\@])>;
@@ -583,7 +605,7 @@ sub lingy_expr($expr) {
     my @groups;
     while (@$tokens) {
         push @groups, eval { $self->group };
-        die "Failed to parser expr '$expr': '$@'" if $@;
+        die "Failed to parse expr '$expr': '$@'" if $@;
     }
     join ' ', map {
         ref($_) ? $self->group_print($_) : $_;
@@ -616,8 +638,11 @@ sub group_list($s, $l) {
     my $group = $s->group_rest;
     return $group if $l or @$group != 3 or $group->[1] !~ qr<^$op$>;
 
+    my $oper = $group->[1];
+    $oper = '-range' if $oper eq '..';
+
     # TODO Support infix group > 3
-    [ $group->[1], $group->[0], $group->[2] ];
+    [ $oper, $group->[0], $group->[2] ];
 }
 
 sub group_call($s, @t) {
@@ -752,6 +777,7 @@ sub o {
 
 sub tag_node($n) { o;
     return 1 if $n->{ytag};
+    $n = transform($n);
     if (is_map($n)) {
         for my $p (pairs($n)) {
             tag_catch($p) or
@@ -763,6 +789,7 @@ sub tag_node($n) { o;
             tag_let($p) or
             tag_loop($p) or
             tag_try($p) or
+            tag_when($p) or
             tag_call($p) or
             XXX $p, "Unable to implicitly tag this map pair.";
         }
@@ -781,6 +808,30 @@ sub tag_node($n) { o;
     1;
 }
 
+sub transform($n) {
+    if (is_map($n)) {
+        for my $p (pairs($n)) {
+            my ($k, $v) = @$p;
+
+            $k->{text} =
+                "$k" eq '???' ? 'cond' :
+                "$k" eq '^^^' ? 'recur' :
+                $k->{text};
+
+            if ("$k" eq 'cond' and is_map($v)) {
+                $p->[1] = bless {
+                    elem => [
+                        map { delete($_->{xkey}); $_ }
+                        map { @$_ } @{$v->{pair}}
+                    ],
+                }, 'seq';
+            }
+        }
+    }
+
+    return $n;
+}
+
 sub tag_map($n) { o;
 }
 
@@ -797,8 +848,7 @@ sub tag_val($n) { o($n);
     } elsif (is_plain($n)) {
         is_key($n) or
         tag_scalar($n) or
-        tag_ysexpr($n) or
-        tag_error("Unresolvable plain scalar");
+        tag_ysexpr($n);
     } else {
         tag_str($n);
     }
@@ -806,10 +856,18 @@ sub tag_val($n) { o($n);
 
 sub tag_call($p) {
     my ($k, $v) = @$p;
-    if ($k =~ /^$sym$/) {
+    if ($k =~ /^$sym($bp?)$/) {
+        my $args = $1;
         $k->{ytag} =
             "$k" eq 'use'
                 ? "$k" :'call';
+
+        # Empty (null) value
+        if (is_plain($v) and text($v) eq '') {
+            err "Use 'foo():' for a call with no args"
+                if $args eq '';
+        }
+
         tag_node($v);
     }
 }
@@ -827,14 +885,14 @@ sub tag_def($p) {
 
 sub tag_defn($p) {
     my ($k, $v) = @$p;
-    return unless $k =~ /^$sym$bp$/;
+    return unless $k =~ /^(?:defn|defmacro)\s+$sym$bp$/;
     $k->{ytag} = 'defn';
     tag_node($v);
 }
 
 sub tag_defn_multi($p) {
     my ($k, $v) = @$p;
-    return unless $k =~ /$sym/ and is_map($v);
+    return unless $k =~ /^(?:defn|defmacro)\s+$sym$/ and is_map($v);
     for my $p (pairs($v)) {
         return unless $p->[0] =~ /^$bp$/;
     }
@@ -859,7 +917,7 @@ sub tag_istr($n) {
 
 sub tag_fn($p) {
     my ($k, $v) = @$p;
-    return unless $k =~ /^$bp$/;
+    return unless $k =~ /^fn\s+$bp$/;
     $k->{ytag} = 'fn';
     tag_node($v);
 }
@@ -898,8 +956,15 @@ sub tag_try($n) {
     $n->{ytag} = 'try' if $n =~ /^try$/;
 }
 
+sub tag_when($p) { o;
+    my ($k, $v) = @$p;
+    return unless $k =~ /(?:\)|. )[?|]$/;
+    $k->{ytag} = 'when';
+    tag_node($v);
+}
+
 sub tag_ysexpr($n) {
-    $n->{text} =~ s/^\\//;
+    $n->{text} =~ s/^\.(?!\d)//;
     $n->{ytag} = 'ysexpr';
 }
 
@@ -953,9 +1018,11 @@ sub tag_ysexpr($n) {
         my $self = bless {
             elem => [@elems],
         }, $class;
-        $refs{$event->{anch}} = $self
-            if $event->{anch} ne '-';
-        $events{Scalar::Util::refaddr($self)} = $event;
+        if ($event) {
+            $refs{$event->{anch}} = $self
+                if $event->{anch} ne '-';
+            $events{Scalar::Util::refaddr($self)} = $event;
+        }
         return $self;
     }
     sub add {
@@ -977,15 +1044,17 @@ sub tag_ysexpr($n) {
     );
     sub new {
         my ($class, $event, $text) = @_;
-        $text //= $event->{valu};
+        $text //= $event->{valu} // '';
         $text =~ s/\\([nt\\\"])/$escapes{$1}/g;
         my $self = bless {
             text => $text,
         }, $class;
-        delete $event->{valu};
-        $refs{$event->{anch}} = $self
-            if $event->{anch} ne '-';
-        $events{Scalar::Util::refaddr($self)} = $event;
+        if ($event) {
+            delete $event->{valu};
+            $refs{$event->{anch}} = $self
+                if $event->{anch} ne '-';
+            $events{Scalar::Util::refaddr($self)} = $event;
+        }
         return $self;
     }
 }
